@@ -92,6 +92,9 @@ int main(int argc, char *argv[])
   struct utimbuf utb;
   char *remote_unlink = NULL;
 
+#ifdef SOCKS
+  SOCKSinit(argv[0]);
+#endif
   term = -1;
   fd = -1;
   first = -1;
@@ -149,7 +152,7 @@ int main(int argc, char *argv[])
 
     /* close the output file if it was left open (it shouldn't be) */
     if (fd > 0) {
-      close(fd);
+      x__close(fd);
       fd = -1;
     }
 
@@ -486,7 +489,7 @@ int main(int argc, char *argv[])
 
     /* Reopen the term connection */
     if (term >= 0)
-      close(term);
+      x__close(term);
     if ((term = socket_connect_server(-1, term_server)) < 0) {
       fprintf(stderr,"Term: %s\n", command_result);
       exit(1);
@@ -517,7 +520,7 @@ int main(int argc, char *argv[])
     }
     else {
       int err = 0;
-      close(fd);
+      x__close(fd);
       fd = -1;
       if (stat(local, &st)) {
         fprintf(stderr, "\n\tCan't stat local file after download!\n");
