@@ -30,9 +30,6 @@
 				/* doesn't use 30-40% of the cpu time. */
 				/* Note: This # is really just accurate to */
 				/* an order of magnitude. */
-#ifndef CLOCKS_PER_SEC 
-#define CLOCKS_PER_SEC 1000
-#endif
 #define FUDGE_FACTOR 3
 
 /* truly global vars. Used everywhere. */
@@ -40,7 +37,7 @@ unsigned long current_time = 0;
 int do_shutdown	 = 0;
 int term_debug = 0;
 int compressing = 1;
-struct Onetime onetime_term[MAX_CLIENTS];
+struct Onetime onetime_term[MAX_CLIENTS+1];
 /*------------------------------------------------------------------------*/
 /* Various global variables. Note that most of these vars are local to    */
 /* this file. Where possible (i.e. where it doesn't impinge on efficiency) */
@@ -52,6 +49,7 @@ int num_clients = 0;
 unsigned long baudrate = 1;
 int max_cps = 0;
 int auto_retrain = 0; /* Don't automatically adjust timing values */
+int use_term_socket = 1; /* This can be deactivated for a one-way term */
 int bytes_left = _POSIX_PIPE_BUF; /* bytes available to send now.. */
 int fudge_flow = 0; /* should we generate periodic control-Q's */
 int byte_shift = 0; /* So we can map less frequently uses section down to */
@@ -85,6 +83,7 @@ int user_share = -1;
 int modem_noise = -1;
 static int hangup_on_exit = 0;	/* tell local modem to hang up on exit */
 static int explicit_hangup = 1;	/* local modem needs ath0 after +++ to hangup */
+static int frequency; 	/* This determines the minimum delay frequency */
 
 int stat_modem_in = 0,
   stat_modem_recv = 0,
@@ -99,12 +98,17 @@ char *home_p = NULL;
 
 int rshtype = 0;  /* Default to allow rsh and use login shell_command */
 
-static char ownroot[256] = "";
-static char shell_command[256] = "";
+static char *private_path = NULL;
+static char *private_path_end = NULL;
+static char socket_path[PATH_MAX] = "\0";
+static char ownroot[PATH_MAX] = "\0";
+static char shell_command[PATH_MAX] = "\0";
 char escapes[256];
 char ignores[256];
 
-int modem_in = 0, modem_out = 1;
+unsigned int term_port = 0;
+
+int modem_in = 0, modem_out = 1, modem_sock = -1;
 /*------------------------------------------------------------------------*/
 /* module function prototypes */
 
@@ -120,12 +124,15 @@ void clear_buffers(struct Client *cl);
 /*------------------------------------------------------------------------*/
 /* main. */
 
+
 void read_int_type(int *o, char *c) {
 
 	/* This reads an integer with "on" == 1, "off" == 0, and default == 1 */
 
   if (isdigit(*c))
     *o = atoi(c);
+  else if (! strncmp(c, "tcp", 3)) 
+    *o = 2;
   else
     *o = !!strncmp(c, "off", 3);
 }
@@ -139,18 +146,29 @@ void read_ulong_type(unsigned long *o, char *c) {
   else
     *o = !!strncmp(c, "off", 3);
 }
-        
+       
 void var_init(void) {
   int k;
+  extern int unix_domain;
 
+#ifdef TERM_NFS_DIR
+  unix_domain = 0;
+#else
+  if (use_term_socket == 2) unix_domain = 0;
+#endif
+
+  frequency = CLK_TCK;
   get_term_localaddr(inet_addr("127.0.0.1"));
   memset(term_remotehost, 0, sizeof(term_remotehost));
   strcpy(term_remotehost,"remotehost");
 
-  for(k=0;k<MAX_CLIENTS;k++) {
+  for(k=0;k <= MAX_CLIENTS;k++) {
     onetime_term[k].socket = -1;
     onetime_term[k].timeout = 0;
+    onetime_term[k].owner = -1;
+    onetime_term[k].errors = 0;
   }
+  if (! share) onetime_term[MAX_CLIENTS].owner = getuid();
 
   if (fudge_flow < 0)
     fudge_flow = 0;
@@ -167,14 +185,12 @@ void var_init(void) {
       breakout_char,breakout_char,breakout_char,breakout_char,breakout_char);
 
   if ( seven_bit_out ) {
-    extern int tok_byte_mask_out, tok_byte_width_out;
     out_mask = 127;
     tok_byte_mask_out = 127;
     tok_byte_width_out = 7;
   }
 
   if (seven_bit_in) {
-    extern int tok_byte_mask_in, tok_byte_width_in;
     in_mask = 127;
     tok_byte_mask_in = 127;
     tok_byte_width_in = 7;
@@ -189,7 +205,7 @@ void var_init(void) {
     baudrate = 0;
 
   max_cps = baudrate ? (2 * baudrate) /
-    (unsigned long) ((int)(0.5 + 2 * stop_bits) + (seven_bit_out ? 14 : 16))
+    (unsigned long) ((int)(0.5 + 2 * stop_bits) + (tok_byte_width_out << 1))
     : 0;  /* avoiding round-off problems */
 
   if ( block_size < 0) block_size = (max_cps) ? 512 : PIPE_BUFFER;
@@ -215,16 +231,16 @@ void var_init(void) {
  
   if ( term_inc < 20 ) term_inc = 20;
 
-  if ( term_inc < FUDGE_FACTOR * (max_cps / block_size)) {
+  if ( term_inc < FUDGE_FACTOR * (max_cps / (2 * packet_len + 20))) {
     k = term_inc;
-    term_inc = FUDGE_FACTOR * (max_cps / block_size); 
+    term_inc = FUDGE_FACTOR * (max_cps / (2 * packet_len + 20)); 
     if (k == 20) packet_timeout = (packet_timeout * term_inc) / k;
   }
 
   if ( delay > 400000 / (unsigned long) term_inc)
     delay = 400000 / (unsigned long) term_inc;
 
-  if ( 2 * term_inc > CLOCKS_PER_SEC ) delay = 0;
+  if ( max_cps / (packet_len + 10) > 0.9 * frequency ) delay = 0;
 
   if (window_size_max < 0 ) {
     window_size_max = 2 + (max_cps / 199);
@@ -257,6 +273,11 @@ void main_init(void) {
 				/* In case of sudden death, do some */
 				/* minimal clean up. */
 void sig_quit(int dummy) {
+  if (use_term_socket) unlink(socket_path);
+  if (private_path) {
+    *private_path_end = 0;
+    rmdir(private_path);
+  }
   set_block(0);
   set_block(1);
   terminal_restore(0,0);
@@ -267,6 +288,11 @@ void sig_quit(int dummy) {
 }
 				/* Drop a core. */
 void sig_core(int dummy) {
+  if (use_term_socket) unlink(socket_path);
+  if (private_path) {
+    *private_path_end = 0;
+    rmdir(private_path);
+  }
   set_block(0);
   set_block(1);
   terminal_restore(0,0);
@@ -275,6 +301,16 @@ void sig_core(int dummy) {
   set_block(modem_out);
   abort();
 }
+
+				/* This closes the modem connection. */
+void sig_hup(int dummy) { 
+  if (term_port < 1)
+    sig_quit(0);
+  if (modem_in >= 0) 
+    x__close(modem_in);
+  modem_in = modem_out = -1;
+}
+
 				/* We ignore this signal. There was */
 				/* some problem with it on linux. A */
 				/* bit odd, and I can't track it down. */
@@ -325,8 +361,9 @@ void read_rc(char *p) {
 #define FLOAT_TYPE 5
   static char int_type=INT_TYPE, char_type=CHAR_TYPE, range_type=RANGE_TYPE,
 	ulong_type=ULONG_TYPE, float_type=FLOAT_TYPE;
-#define RC_OPTIONS_SIZE 90
-  int m, n;
+#define RC_OPTIONS_SIZE 96
+  int m, n, i;
+  char *where="/.term";
   static char *rc_options[RC_OPTIONS_SIZE] = {
 	"compress",	& int_type,	(char *) &compressing,
 	"breakout",	& int_type,	(char *) &breakout_char,
@@ -357,7 +394,9 @@ void read_rc(char *p) {
 	"packetsize",	& int_type,	(char *) &packet_len,
 	"terminate",	& char_type,	(char *) breakout_string,
 	"hangup_on_exit", & int_type,	(char *) &hangup_on_exit,
-	"explicit_hangup", & int_type,	(char *) &explicit_hangup
+	"explicit_hangup", & int_type,	(char *) &explicit_hangup,
+	"socket",	& int_type,	(char *) &use_term_socket,
+	"frequency",	& int_type,	(char *) &frequency
       };
 
 #define DEFAULT_PATH "/usr/local/lib/term:/usr/lib/term:/usr/etc:/etc" 
@@ -374,16 +413,20 @@ void read_rc(char *p) {
   else if (savedeid >= 0)
     setgid(savedeid);
 
-  do {
-    if ((file = get_term_path(&ptr))) {
-      if(share != 1) 
-        strcat(file, "/.term");
-      strcat(file, "/termrc");
-      if (p != NULL) if ( p[0] )
-        sprintf(&file[strlen(file)], ".%s",p);
-      if (! eaccess(file, R_OK)) break;
-    }else break;
-  } while ( ptr );
+  for(i=!! share;! file && i < 2 ; ++i) {
+    ptr = NULL;
+    if (i == 1) where = NULL;
+    do {
+      if ((file = get_term_path(&ptr))) {
+        if (where)
+          strcat(file, where);
+        strcat(file, "/termrc");
+        if (p != NULL) if ( p[0] )
+          sprintf(&file[strlen(file)], ".%s",p);
+        if (! eaccess(file, R_OK)) break;
+      }else break;
+    } while ( ptr );
+  }
 
   if (file) {
     if (!quiet) 
@@ -532,7 +575,7 @@ void read_rc(char *p) {
 #endif
 }
 
-#define GETOPT_OPTS "Aac:d:f:l:n:ors:t:v:w:P:S:q"
+#define GETOPT_OPTS "Aac:d:f:l:n:ors:t:u:v:w:p:P:S:q"
 
 int main(int argc, char *argv[]) 
 {
@@ -541,6 +584,10 @@ int main(int argc, char *argv[])
   char *ptr=NULL, *path=NULL;
   int umask_old;
 
+#ifdef SOCKS
+  SOCKSinit(argv[0]);
+#endif
+
   breakout_string[0] = 0;
   term_opterr = 0;
   while ((c = term_getopt (argc, argv, GETOPT_OPTS)) != EOF) 
@@ -601,13 +648,7 @@ int main(int argc, char *argv[])
       break;
     case 'a':
       {
-	extern int tok_byte_mask_out, tok_byte_width_out,
-	  tok_byte_mask_in, tok_byte_width_in;
-	tok_byte_mask_out = tok_byte_mask_in = 127;
-	tok_byte_width_out = tok_byte_width_in = 7;
-
 	seven_bit_in = seven_bit_out = 1;
-	in_mask = out_mask = 127;
 	break;
       }
     case 'f': 
@@ -616,35 +657,39 @@ int main(int argc, char *argv[])
     case 'l':
       i = open(term_optarg, O_RDWR | O_CREAT | O_TRUNC, 0644);
       if (i < 0) {
-	perror("open");
+	x__perror("open");
 	fprintf(stderr, "Unable to open log file %s\r\n", term_optarg);
 	break;
       }else {
         modem_noise = 0;
       }
       if (i != 2) {
-	close(2);		/* Just to make sure.. */
-	dup2(i, 2);
-	close(i);
+	x__close(2);		/* Just to make sure.. */
+	x__dup2(i, 2);
+	x__close(i);
       }
       break;
+    case 'p':
+      term_port=atoi(term_optarg);
+      if(modem_in >= 0) x__close(modem_in);
+      if(modem_out >= 0) x__close(modem_out);
+      modem_in = modem_out = -1;
+      break;
     case 'r': remote = 1; break;
     case 't':
       read_ulong_type(&packet_timeout,term_optarg);
       break;
     case 'v':
-
 				/* add a device to the list. */
       i = open(term_optarg, O_RDWR);
       if (i < 0) {
-	perror("open");
+	x__perror("open");
 	fprintf(stderr,"Unable to open modem device %s\r\n", term_optarg);
 	break;
       }
-      if(modem_in>=0) close(modem_in);
-      if(modem_out>=0) close(modem_out);
-      modem_in = i;
-      modem_out = i;
+      if(modem_in>=0) x__close(modem_in);
+      if(modem_out>=0) x__close(modem_out);
+      modem_in = modem_out = i;
       break;
     case 'w':
       read_int_type(&window_size_max,term_optarg);
@@ -662,6 +707,9 @@ int main(int argc, char *argv[])
     case 's':
       read_ulong_type(&baudrate,term_optarg);
       break;
+    case 'u':
+      read_int_type(&use_term_socket,term_optarg);
+      break;
     case 'q':
     case 'S':
       break;
@@ -680,12 +728,21 @@ int main(int argc, char *argv[])
   if (term_server[0])
     read_rc(term_server);
 
-  if (!share) {
+  { char tmp[1024];
+    sprintf(tmp,"TERMSERVER=%s",term_server);
+    term_putenv(tmp);
+  }
+
+  if (! use_term_socket) {
+    DEBUG_FP(stderr,"WARNING: Term will be inaccessible on this side");
+    s = -1;
+  }else if (!share) {
     do {
       if ((path = get_term_path(&ptr))) {
         strcat(path,"/.term");
+        mkdir(path,0700);
         if (access(path, X_OK | R_OK | W_OK) < 0) continue;
-        if (chmod(path,0700)  < 0) {
+        if ((chmod(path,0700)  < 0)) {
           if (!quiet) 
             fprintf(stderr,"Can't set permissions on %s to 0700\r\n",
               path);
@@ -694,7 +751,7 @@ int main(int argc, char *argv[])
         break;
       }else break;
     } while ( ptr );
-    if (! path) {
+    if (use_term_socket && ! path) {
       fprintf(stderr, "Failed to find any of:\r\n");
       ptr = NULL;
       do {
@@ -762,26 +819,27 @@ int main(int argc, char *argv[])
     strcat(path,term_server);
 
 /* Remove any old sockets */
-    i=strlen(path);
+    private_path = path;
+    private_path_end = &path[strlen(private_path)];
     strcat(path,"/socket");
     unlink(path);
-    path[i]='\0';
 
 /* Remove and recreate the private directory, so we can be sure we can see it */
-    rmdir(path);
+    *private_path_end = 0;
+    rmdir(private_path);
     if (share == 2) {
-      mkdir(path, 0700|S_ISUID);
-      chmod(path, 0700|S_ISUID);
+      mkdir(private_path, 0700|S_ISUID);
+      chmod(private_path, 0700|S_ISUID);
     }else {
-      mkdir(path, 0770|S_ISGID);
-      chmod(path, 0770|S_ISGID);
-    };
+      mkdir(private_path, 0770|S_ISGID);
+      chmod(private_path, 0770|S_ISGID);
+    }
   
 /* Finally make sure we can access the directory */
 
-    if (eaccess(path, getuid() ? (X_OK | W_OK | R_OK) : 0) < 0) {
+    if (eaccess(private_path, getuid() ? (X_OK | W_OK | R_OK) : 0) < 0) {
       fprintf(stderr, "Can't read, write, &/or execute %s\r\n",
-        path);
+        private_path);
       exit(1);
     }
 
@@ -791,25 +849,31 @@ int main(int argc, char *argv[])
     umask(umask_old);
   }
 
+  /* initialize variables */
+
+  var_init();
+
 /* It doesn't hurt to unlink twice ... */  
-  unlink(path);
-  if ((s = bind_unix(path)) < 0) 
-    exit(1);
+  if (use_term_socket) {
+    unlink(path);
+    strcpy(socket_path,path);
+    if ((s = bind_unix(path)) < 0) exit(1);
+  }else {
+    s = -1;
+  }
 
   setuid(getuid());
   setgid(getgid());
 
   /* init modules */
  
-  var_init();
-
   serial_init();
   compress_init();
   update_time();
 
   if (ownroot[0]) {
-    if (chroot(ownroot) < 0) {
-      perror("chroot");
+    if (x__chroot(ownroot) < 0) {
+      x__perror("chroot");
       if (rshtype > 0) {
         fprintf(stderr, "Your chroot failed.  Exiting rather than risking invasion\n");
         exit(1);
@@ -826,32 +890,57 @@ int main(int argc, char *argv[])
 #else
   signal(SIGCHLD, sig_child);
 #endif /* SYSV */
-  signal(SIGHUP, sig_quit);
+  signal(SIGHUP, sig_hup);
   signal(SIGPIPE, sig_ignore);
   signal(SIGINT, sig_quit);
   signal(SIGQUIT, sig_quit);
   signal(SIGIOT, sig_core);
   signal(SIGSEGV, sig_core);
   signal(SIGALRM, sig_ignore);
+  signal(SIGBUS, sig_ignore);
 
   terminal_save(0);
-  set_nonblock(s);
-  set_nonblock(modem_in);
-  set_nonblock(modem_out);
-  terminal_raw(modem_in);
-  terminal_raw(modem_out);
+  if (s >= 0) set_nonblock(s);
+
+  if (modem_in >= 0 || modem_out >= 0 ) {
+#ifndef NO_TTYNAME
+	/* Some OS's insist that streams must be opened with O_RDWR... */
+    if (modem_in != modem_out) {
+      int t;
+      if (modem_in >= 0 && isatty(modem_in) &&
+          (t = open(ttyname(modem_in),O_RDWR)) >= 0 ) {
+        dup2(t,modem_in);
+        close(t);
+      }
+      if (modem_out >= 0 && isatty(modem_out) &&
+          (t = open(ttyname(modem_out),O_RDWR)) >= 0 ) {
+        dup2(t,modem_out);
+        close(t);
+      }
+    }
+#endif
+    set_nonblock(modem_in);
+    set_nonblock(modem_out);
+    terminal_raw(modem_in);
+    terminal_raw(modem_out);
+  }
 
   main_loop(s);
 
+  if (private_path) {
+    *private_path_end = 0;
+    rmdir(private_path);
+  }
   	/* with C_QUIT 2 or if hangup_on_exit and not C_QUIT 1 */
   hangup = do_shutdown == 2 || (hangup_on_exit && do_shutdown!=1);
 
+  if (use_term_socket) unlink(path);
   set_block(modem_in);
   set_block(modem_out);
   terminal_restore(0,hangup);
   terminal_restore(1,hangup);
 
-  set_block(s);
+  if (s >= 0) set_block(s);
 
 	/* If terminal_restore didn't hangup, this might. */
 	/* This is way too specific for my blood.  This assumes your modem */
@@ -871,7 +960,7 @@ int main(int argc, char *argv[])
 /*-----------------------------------------------------------------------*/
 void check_client(int cl, int ret) {
   DEBUG_MAIN(stderr, "%s: termerrno == %d\n", term_server, termerrno);
-  if (!termerrno) return;
+  if (!termerrno && checkfd(clients[cl].fd) >= 0) return;
   if (clients[cl].state == 3 ) {
     DEBUG_FP(stderr, "%s:truncating out_buff\n", term_server);
     clients[cl].out_buff.size = 0;
@@ -880,16 +969,16 @@ void check_client(int cl, int ret) {
   }
 #if 0
   if (ret < 0)
-    perror("client gave this");
+    x__perror("client gave this");
 #endif
 				/* Ok. Close the descriptor. */
-  if(clients[cl].fd >= 0) close(clients[cl].fd);
+  if(clients[cl].fd) x__close(clients[cl].fd);
   clients[cl].fd = -1;
 				/* And go to state 2. */
   clients[cl].state = 2;
 }
 
-int new_client(int fd) {
+int new_client(int fd, int uid) {
   int j;
 
   for (j = remote; j < MAX_CLIENTS;j += 2)
@@ -916,8 +1005,9 @@ int new_client(int fd) {
   clients[j].name[0] = 0;
   clients[j].parent = -1;
   clients[j].peername.sin_port = ~0;
-  clients[j].peername.sin_addr.s_addr = term_localaddr;
+  clients[j].peername.sin_addr.s_addr = htonl(term_localaddr);
   clients[j].peername.sin_family = ~0;
+  clients[j].owner = uid;
   set_nonblock(fd);
   return j;
 }
@@ -925,13 +1015,46 @@ int new_client(int fd) {
 /* Main loop. Hangs around waiting for things to get ready, and calls to  */
 /* appropriate routines. 						  */
 
-void main_loop(int socket) {
+void main_loop(int sock) {
   struct timeval timeout;
   fd_set reads, writes, excepts;
   int i, j, k, empty = 1, csocket = -1, bytes_received = 0, script_fd = -1;
-  int max = 0;
+  int max;
+  int on = 1;
+  int cerrors = 0, select_error = 0;
 
   while (!do_shutdown) {
+    max = -1;
+
+/* First thing we do is make sure we have a modem connection */
+    if (modem_in < 0 || modem_out < 0 || checkfd(modem_in) < 0) {
+      if (term_port < 1) {
+	exit(1);
+      }else {
+        struct sockaddr_in addr_in;
+        int din = sizeof(addr_in);
+        if(modem_in>=0) x__close(modem_in);
+        modem_in = modem_out = -1;
+        if ((modem_sock=bind_tcp_listen(term_port,1)) < 0) {
+          if (modem_sock == -1) {
+            fprintf(stderr,"Port %u is already bound\n",term_port);
+          }else {
+            fprintf(stderr, "Can't listen to port %u\n",term_port);
+          }
+          exit(1);
+        }
+        fprintf(stderr, "Listening to port %u for a modem connection\n",term_port);
+        while ((modem_in = modem_out = 
+          x__accept(modem_sock, (struct sockaddr *) &addr_in, &din)) < 0)
+          x__perror("Accept from modem");
+        set_nonblock(modem_in);
+        setsockopt(modem_sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
+        x__close(modem_sock);
+        modem_sock = -1;
+        setsockopt(modem_in, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
+        fprintf(stderr,"Accepted modem connection\n");
+      }
+    }
 
 /* If the serial out buffer is empty, try and put something in it */
     if (bytes_left > 0 && serial_out.size < bytes_left) {
@@ -944,7 +1067,7 @@ void main_loop(int socket) {
 	do_serial_out(0);
       DEBUG_SER(stderr, "%s:  serial out size %d\n",
 	       term_server, serial_out.size);
-    }
+    } 
 
 /* Set up client stuff */
 /* We select to read if: */
@@ -973,7 +1096,7 @@ void main_loop(int socket) {
         for(l=0;l<MAX_CLIENTS;++l)
           if(clients[l].parent == i && clients[l].state == 1 )
             clients[l].state = 3;
-	if (clients[i].fd > 0) close(clients[i].fd);
+	if (clients[i].fd > 0) x__close(clients[i].fd);
 	clients[i].fd = -1;
 	if (clients[i].state == 3)
 	  clients[i].state = -1; 
@@ -983,36 +1106,45 @@ void main_loop(int socket) {
 				/* If it's a file, we don't need to */
 				/* select() on it. */
       if (clients[i].cl_type == CL_FILE) continue;
+      if (select_error) check_client(i,0);
       if (clients[i].fd < 0) continue; /* If it's not a file, and */
 				       /* there is no fd, then no */
 				       /* select(). */
       if (clients[i].in_buff.size < 2 * packet_len &&
             (clients[i].type & T_RDFILE) && clients[i].state == 1) {
 	FD_SET(clients[i].fd, &reads);
+	if(max < clients[i].fd) max = clients[i].fd;
       }
       if ((clients[i].out_buff.size && (clients[i].type & T_WRFILE)) ||
 	  (clients[i].state == 5)) {
 	FD_SET(clients[i].fd, &writes);
+	if(max < clients[i].fd) max = clients[i].fd;
       }
-      FD_SET(clients[i].fd, &excepts);
-      if (max < clients[i].fd) max = clients[i].fd;
+      if(FD_ISSET(clients[i].fd, &reads)||FD_ISSET(clients[i].fd, &writes))
+        FD_SET(clients[i].fd, &excepts);
     }
     
 /* Select for socket and modem */
 /* We select read for socket if we aren't at the maximum number */
 /* of clients */
-    if (num_clients < MAX_CLIENTS) {
-      FD_SET(socket, &reads);
-      if (max < socket) max = socket;
-      for (k=0;k < MAX_CLIENTS;k++) if(onetime_term[k].socket >= 0){
+    if (sock >= 0 && num_clients < MAX_CLIENTS 
+        && (! select_error || (sock=checkfd(sock)) >= 0)) {
+      FD_SET(sock, &reads);
+      FD_SET(sock, &excepts);
+      if (max < sock) max = sock;
+      for (k=0;k < MAX_CLIENTS;k++) if(onetime_term[k].socket >= 0
+        && (! select_error 
+           || (onetime_term[k].socket=checkfd(onetime_term[k].socket)) >= 0)) {
         FD_SET(onetime_term[k].socket, &reads);
-        if (max < onetime_term[k].socket) max = onetime_term[k].socket;
+        FD_SET(onetime_term[k].socket, &excepts);
+	if (max < onetime_term[k].socket) max = onetime_term[k].socket;
       }
     }
 
 /* We select for read on the modem if the serial in buffer is empty */
     if (!serial_in.size) {
       FD_SET(modem_in, &reads);
+      FD_SET(modem_in, &excepts);
       if (max < modem_in) max = modem_in;
     }
 
@@ -1022,7 +1154,8 @@ void main_loop(int socket) {
       DEBUG_SER(stderr, "%s: dso: fd set\n",
 	       term_server);
       FD_SET(modem_out, &writes);
-      if (max < modem_out) max = modem_out;
+      FD_SET(modem_out, &excepts);
+      if (max < modem_out)  max = modem_out;
     }
 
 /* We need to call the script from the main loop, since we want term up */
@@ -1033,33 +1166,36 @@ void main_loop(int socket) {
 
     if (shell_command[0] && remote_term_version) { 
       script_fd = open_socket(shell_command); 
-        set_nonblock(script_fd);
+      set_nonblock(script_fd);
       shell_command[0] = 0;
     }
-    if (script_fd >= 0) 
+    if (script_fd >= 0 
+      && (! select_error || (script_fd=checkfd(script_fd)) >= 0)) {
       FD_SET(script_fd, &reads);
-
-				/* This is only for my own purposes. */
+      if (max < script_fd) max = script_fd;
+    }
 
 /* This is intended as a short delay to help decrease CPU usage. */
     timeout.tv_sec = 0;
     timeout.tv_usec = (unsigned long int) delay;
-    if(max_cps > 0) select(0, NULL, NULL, NULL, &timeout); 
+    if(max_cps > 0 || max < 0) select(0, NULL, NULL, NULL, &timeout); 
 
 /* Set the timeout value for select(). */
     timeout.tv_sec = 0;
     timeout.tv_usec = 500000; /* 0.5 seconds */
 
 /* do select() */	
-    if (select(max+1, &reads, &writes, &excepts, &timeout) < 0) {
+    if (max > -1) if (select(max+1, &reads, &writes, &excepts, &timeout) < 0) {
 				/* This is perfectly normal. Things */
 				/* that send signals will cause select */
 				/* to exit with an error. */
-#if 0
-      perror("select");
-#endif
+
+      if (++cerrors < 10) 
+        x__perror("select");
+      select_error = 1;
       continue;
     }	
+    select_error = 0;
 
 /* Update current_time. This is maintained in 20th s of a second */
     update_time();
@@ -1068,11 +1204,12 @@ void main_loop(int socket) {
 
 /* See if it is just the script */
 
-    if (script_fd >= 0 && FD_ISSET(script_fd, &reads)) {
+    if (script_fd >= 0 &&
+       (FD_ISSET(script_fd, &reads)||FD_ISSET(script_fd, &excepts))) {
       memset(shell_command,0,sizeof(shell_command));
       j = read(script_fd, &shell_command[1], sizeof(shell_command)-1);
       if ( j <= 0 && errno != ERR_BLOCK) {
-        close(script_fd);
+        x__close(script_fd);
         script_fd = -1;
       }else {   /* I copy this exactly to stderr */
         if (write_noise < 2) fwrite(&shell_command[1],j,sizeof(char),stderr);
@@ -1081,11 +1218,15 @@ void main_loop(int socket) {
 
 /* Can we read from modem  ?? */
     bytes_received = 0;
-    if (!serial_in.size && FD_ISSET(modem_in , &reads)) {
+    if (!serial_in.size && 
+      (FD_ISSET(modem_in, &reads) || FD_ISSET(modem_in, &excepts))) {
       j = read_into_buff(modem_in, &serial_in, 0);
-      if ( j < 0 && errno != ERR_BLOCK) {
-	perror("read from modem");
-	return;
+      if ( (j < 0 || (!j && termerrno && term_port > 0)) && 
+        (errno != ERR_BLOCK || term_port > 0)) {
+	x__perror("read from modem");
+	x__close(modem_in);
+	modem_in = modem_out = -1;
+	continue;
       }
       else{
         stat_modem_in += j;
@@ -1100,15 +1241,15 @@ void main_loop(int socket) {
 
 /* test for new client */
 
-    for(k=0;k <= MAX_CLIENTS;k++){
+    if (sock >= 0) for(k=0;k <= MAX_CLIENTS;k++){
       switch (k){
         case MAX_CLIENTS:
-          csocket = socket;
+          csocket = sock;
           break;
         default:
           if (onetime_term[k].socket > 0 && 
               current_time > onetime_term[k].timeout) {
-            close(onetime_term[k].socket);
+            x__close(onetime_term[k].socket);
             onetime_term[k].socket = -1;
           }
           csocket = onetime_term[k].socket;
@@ -1118,6 +1259,7 @@ void main_loop(int socket) {
       if (FD_ISSET(csocket, &reads)) { /* try for a connect. */
 
         struct sockaddr_in addr_in;
+        int uid = -1;
 
 #if defined(STREAMS_PIPE) || defined(X_STREAMS_PIPE)
  	i = CheckClientConnection(csocket);
@@ -1130,46 +1272,62 @@ void main_loop(int socket) {
 #endif
         {
           int din = sizeof(addr_in);
-          if ((i = accept(csocket , (struct sockaddr *) &addr_in, &din)) >= 0) {
+          if ((i = x__accept(csocket , (struct sockaddr *) &addr_in, &din)) >= 0) {
             if (addr_in.sin_family == AF_INET) {
-              if (addr_in.sin_addr.s_addr != htonl(term_localaddr)
-                  && addr_in.sin_addr.s_addr != INADDR_ANY 
-                  && addr_in.sin_addr.s_addr != inet_addr("127.0.0.1") ) {
-                close(i);
+              char *user = NULL;
+
+              if (( addr_in.sin_addr.s_addr != htonl(term_localaddr)
+                    && addr_in.sin_addr.s_addr != INADDR_ANY 
+                    && addr_in.sin_addr.s_addr != inet_addr("127.0.0.1") )
+                 ||( (user = getconnname(i)) != NULL
+                    && (uid = getuserid(user)) != onetime_term[k].owner
+                    && uid != -1
+                    && uid != 0
+                    && onetime_term[k].owner != -1 )) {
+                x__close(i);
                 i = -1;
-                DEBUG_FP(stderr,"%s: rejected connection from %s\n",
-                  term_server, inet_ntoa(addr_in.sin_addr));
+                fprintf(stderr,"%s: rejected connection from user %s@%s\n",
+                  term_server, user ? user : "unknown", inet_ntoa(addr_in.sin_addr));
               }else {
-                DEBUG_FP(stderr,"%s: accepted connection %d from %s\n",
-                  term_server, i, inet_ntoa(addr_in.sin_addr));
+                DEBUG_FP(stderr,"%s: accepted connection %s@%s:%u\n",
+                  term_server, user ? user : "unknown", inet_ntoa(addr_in.sin_addr),
+		  (unsigned int) ntohs(addr_in.sin_port));
               }
             }else {
               DEBUG_FP(stderr,"%s: accept connection from unix domain\n",
                 term_server);
             } 
+          }else {
+            x__perror("Accept");
+            if (++onetime_term[k].errors > 100) {
+              if (k < MAX_CLIENTS) x__close(csocket); 
+              onetime_term[k].socket = -1;
+              onetime_term[k].errors = 0;
+            }
+            continue;
           }
         }
         if (k < MAX_CLIENTS) {
 	  int on=1;
           onetime_term[k].socket = -1;
-          close(csocket);
+          x__close(csocket);
 	  setsockopt(i, SOL_SOCKET, SO_REUSEADDR,
             (char *)&on, sizeof(on));
         }
         
         if (i >= 0) { /* a new client */
-	  if((j = new_client(i)) < 0) 
-            close(i);
+	  if((j = new_client(i,uid)) < 0) 
+            x__close(i);
           else
             memcpy(&clients[i].peername,&addr_in,sizeof(clients[i].peername));
         }
         else if (termerrno != 1) {
 #if 0
-	  perror("accept");
+	  x__perror("accept");
 #endif
         }
       }
-    };
+    }
 
 /* test for data being read from clients */
     for (i = 0; i < MAX_CLIENTS;++i) {
@@ -1206,7 +1364,7 @@ void main_loop(int socket) {
       case CL_FILE:
 	if (!(clients[i].type & T_RDFILE)) continue;
 	if (! clients[i].dump_count) {
-          close(clients[i].fd);
+          x__close(clients[i].fd);
           clients[i].fd = -1;
           clients[i].cl_type = CL_SOCKET;
           clients[i].type = T_RDFILE | T_WRFILE;
@@ -1262,14 +1420,14 @@ void main_loop(int socket) {
 	      errno = ETIMEDOUT;
 	      ret_fail(&clients[i], 0, 1, 
                 "connect() timed out");
-	      perror("async connect()");
+	      x__perror("async connect()");
 	      clients[i].state = 2;
 	    }
 	    break;
 	  default:
 	    ret_fail(&clients[i], 0, 1,
               "connect() timed out");
-	    perror("async connect()");
+	    x__perror("async connect()");
 	    clients[i].state = 2;
 	    break;
 	  }
@@ -1313,7 +1471,6 @@ void main_loop(int socket) {
     
 /* Ok. Can we write to modem ??? */
     if (FD_ISSET(modem_out, &writes) && bytes_left > 0) {
-      
       int t = bytes_left;
       if (t > serial_out.size)
 	t = serial_out.size;
@@ -1324,8 +1481,10 @@ void main_loop(int socket) {
       j = write_from_buff(modem_out, &serial_out, t);
       if (j < 1 && termerrno != 1) {
 	errno = termerrno - 1;
-	perror("write to modem");
-	return;
+	x__perror("write to modem");
+	x__close(modem_out);
+	modem_in = modem_out = -1;
+	continue;
       }else {
 	stat_modem_out += j;
         bytes_left -= j + bytes_received + (modem_noise ? modem_noise - 1 : 0);
@@ -1334,7 +1493,7 @@ void main_loop(int socket) {
 
   } /* while loop */
 
-  if (script_fd >= 0) close(script_fd);
+  if (script_fd >= 0) x__close(script_fd);
  
 } /* function */
 
