@@ -94,6 +94,10 @@ int main(int argc, char *argv[])
   int permissions;
   long atime = 0, mtime = 0, rmtime = 0;
 
+#ifdef SOCKS
+  SOCKSinit(argv[0]);
+#endif
+
   term = -1;
   fd = -1;
   first = -1;
@@ -128,7 +132,7 @@ int main(int argc, char *argv[])
   get_from_buffer(&buffer);
 
 #ifdef linux
-  term = socket(AF_INET, SOCK_STREAM, 0);
+  term = x__socket(AF_INET, SOCK_STREAM, 0);
 #else
   term = -1;
 #endif
@@ -165,7 +169,7 @@ int main(int argc, char *argv[])
 
     /* close the input file if it was left open */
     if (fd > 0) {
-      close(fd);
+      x__close(fd);
       fd = -1;
     }
 
@@ -392,7 +396,7 @@ int main(int argc, char *argv[])
       /* do the local file seek */
       if (lseek(fd, (off_t)remote_size, 0) < (off_t)0) {
 	fprintf(stderr, "\tSkipped, local seek failed, ");
-	perror("Reason given");
+	x__perror("Reason given");
         j=1;
 	continue;
       }
@@ -580,7 +584,7 @@ int main(int argc, char *argv[])
       if (( fd && (l != (long)st.st_size) ) ||
          ( (!fd) && (l != bytesent) )) {
        fprintf(stderr, "F\n\tRemote file is a different size from local after"
-               " upload!\n\t(r=%ld,l=%ld)\n", l, st.st_size); 
+               " upload!\n\t(r=%ld,l=%ld)\n", l, (long)st.st_size); 
        j=1;
        continue;
       }
